# -*- coding: utf-8 -*-
"""
Self Attention

Author -- Dhananjay Tomar

"""

import torch


class SelfAttention(torch.nn.Module):
    def __init__(self, embed_dim):
        super().__init__()

        self.embed_dim = embed_dim
        self.query_projection = torch.nn.Linear(embed_dim, embed_dim)
        self.key_projection = torch.nn.Linear(embed_dim, embed_dim)
        self.value_projection = torch.nn.Linear(embed_dim, embed_dim)
        # Optional: initialise a softmax layer here.
        # Otherwise, you may also use torch.nn.functional.softmax directly in the forward function
        self.softmax = torch.nn.Softmax(dim=2)

    def forward(self, x):
        # Step 1: Pass the input through the projection layers to get queries, keys and values
        queries = self.query_projection(x)
        keys = self.key_projection(x)
        values = self.value_projection(x)
        # Step 2: Get the scores, i.e., the dot products between queries and keys
        # queries, keys and values have shape [batch_size, sequence_length, embed_dim]
        # The matrix containing the scores between all pairs of queries and keys
        # should have shape [batch_size, sequence_length, sequence_length]
        # where matrix[batch_index, i, j] = dot_product(q[batch_index, i], k[batch_index, j]) / sqrt(embed_dim)
        # This can be achieved using torch.matmul or torch.bmm (bmm = batch matrix multiply) instead of loops.
        # Think how can one of the 2 matrices can be transformed to get the desired result.
        # Hint: multiplying matrix of shape [a, b] with [b, c] results in a matrix of shape [a, c]
        #       Use .transpose() function of the tensor to swap two dimensions
        scores = torch.bmm(queries, keys.transpose(1, 2)) / (self.embed_dim ** 0.5)
        # Step 3: Get the attention weights
        # Run the matrix either through the softmax layer created in __init__ or use torch.nn.functional.softmax
        # Don't forget to specify the correct dim for softmax.
        attention_weights = self.softmax(scores)
        # Step 4: Weighted average of value vectors
        # Finally, return the weighted average of the value vectors.
        # You can use torch.matmul or torch.bmm again here.
        return torch.bmm(attention_weights, values)


class MultiHeadSelfAttention(torch.nn.Module):
    def __init__(self, embed_dim, num_heads):
        super().__init__()
        assert embed_dim % num_heads == 0, 'Embedding dimension must be a multiple of number of heads.'

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.query_projection = torch.nn.Linear(embed_dim, embed_dim)
        self.key_projection = torch.nn.Linear(embed_dim, embed_dim)
        self.value_projection = torch.nn.Linear(embed_dim, embed_dim)
        self.output_projection = torch.nn.Linear(embed_dim, embed_dim)
        # Optional: initialise a softmax layer here.
        # Otherwise, you may also use torch.nn.functional.softmax directly in the forward function
        self.softmax = torch.nn.Softmax(dim=-1)

    def forward(self, x):
        # Step 1: Pass the input through the projection layers to get queries, keys and values
        queries = self.query_projection(x)
        keys = self.key_projection(x)
        values = self.value_projection(x)
        # Step 2: Split the queries, keys and values into num_heads to
        # have shape [batch, seq_length, num_heads, embed_dim // num_heads]
        # Hint: You just need to use the .reshape() function
        queries = queries.reshape(queries.size(0), queries.size(1), self.num_heads, queries.size(2) // self.num_heads)
        keys = keys.reshape(keys.size(0), keys.size(1), self.num_heads, keys.size(2) // self.num_heads)
        values = values.reshape(values.size(0), values.size(1), self.num_heads, values.size(2) // self.num_heads)
        # Step 3: Compute the scores
        # First, permute Q, K to have shape [batch, num_heads, seq_length, embed_dim // num_heads]
        queries = queries.permute(0, 2, 1, 3)
        keys = keys.permute(0, 2, 1, 3)
        # Now do the matrix multiplication between queries and keys per head to get scores.
        # Hint: You don't need loops. Use torch.matmul instead. And forget to swap dimensions before matmul.
        scores = torch.matmul(queries, keys.transpose(-2, -1)) / (keys.size()[-1] ** 0.5)
        # Step 4: Attention weights
        # Run the matrix either through the softmax layer created in __init__ or use torch.nn.functional.softmax
        attention_weights = self.softmax(scores)
        # Step 5: Weighted average
        # Permute values to have shape [batch, num_heads, seq_length, embed_dim // num_heads] just like above
        values = values.permute(0, 2, 1, 3)
        # Now compute the weighted average
        context_vectors = torch.matmul(attention_weights, values)
        # Concatenate the outputs of different heads
        context_vectors = context_vectors.permute(0, 2, 1, 3)
        s = context_vectors.size()
        context_vectors = context_vectors.reshape(s[0], s[1], s[2] * s[3])
        # Step 6: Run the output through the final linear layer
        return self.output_projection(context_vectors)


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


if __name__ == '__main__':
    batch_size = 128
    seq_len = 100
    embedding_len = 512
    multi_head_att = MultiHeadSelfAttention(embedding_len, 8)
    att = SelfAttention(embedding_len)
    input_x = torch.randn(batch_size, seq_len, embedding_len)
    print(f'Multi-headed self-attention has {count_parameters(multi_head_att)} parameters.')
    print(f'Single-headed self-attention has {count_parameters(att)} parameters.')
    print(f'Output of multi-headed self-attention is of shape {multi_head_att(input_x).shape}')
    print(f'Output of single-headed self-attention is of shape {att(input_x).shape}')
